@@ -1,7 +1,7 @@
 /*----------------------------------------------------------------------------
  ADOL-C -- Automatic Differentiation by Overloading in C++
  File:     ho_rev.c
- Revision: $Id: ho_rev.c 376 2012-12-14 11:22:56Z kulshres $
+ Revision: $Id: ho_rev.c 527 2014-07-15 14:09:31Z kulshres $
  Contents: Contains the routines :
            hos_reverse (higher-order-scalar reverse mode): 
               define _HOS_
@@ -374,7 +374,7 @@ int hov_ti_reverse(
     locint arg1 = 0;
     locint arg2 = 0;
 
-    double coval = 0, *d = 0;
+    double coval = 0;
 
     int indexi = 0,  indexd = 0;
 
@@ -487,6 +487,7 @@ int hov_ti_reverse(
     rp_Atemp  = (revreal *)malloc(k1 * sizeof(revreal));
     rp_Atemp2 = (revreal *)malloc(k1 * sizeof(revreal));
     rp_Ttemp2 = (revreal *)malloc(k * sizeof(revreal));
+    ADOLC_CURRENT_TAPE_INFOS.workMode = ADOLC_HOS_REVERSE;
     /*----------------------------------------------------------------------*/
 #elif _HOV_                                                          /* HOV */
     rpp_A = (revreal**)malloc(ADOLC_CURRENT_TAPE_INFOS.stats[NUM_MAX_LIVES] *
@@ -512,6 +513,7 @@ int hov_ti_reverse(
     rp_Atemp  = (revreal*) malloc(pk1 * sizeof(revreal));
     rp_Atemp2 = (revreal*) malloc(pk1 * sizeof(revreal));
     rp_Ttemp2 = (revreal*) malloc(k * sizeof(revreal));
+    ADOLC_CURRENT_TAPE_INFOS.workMode = ADOLC_HOV_REVERSE;
     /*----------------------------------------------------------------------*/
 #elif _HOS_OV_                                                    /* HOS_OV */
     rpp_A = (revreal**)malloc(ADOLC_CURRENT_TAPE_INFOS.stats[NUM_MAX_LIVES] *
@@ -537,6 +539,7 @@ int hov_ti_reverse(
     rp_Atemp  = (revreal*) malloc(pk1 * sizeof(revreal));
     rp_Atemp2 = (revreal*) malloc(pk1 * sizeof(revreal));
     rp_Ttemp2 = (revreal*) malloc(p*k*sizeof(revreal));
+    ADOLC_CURRENT_TAPE_INFOS.workMode = ADOLC_HOV_REVERSE;
 #endif
     rp_Ttemp  = (revreal*) malloc(k*sizeof(revreal));
     if (rp_Ttemp == NULL) fail(ADOLC_MALLOC_FAILED);
@@ -555,7 +558,7 @@ int hov_ti_reverse(
         fprintf(DIAG_OUT,"\n ADOL-C error: reverse fails because it was not"
                 " preceded\nby a forward sweep with degree>%i,"
                 " keep=%i!\n",degre,degre+1);
-        exit(-2);
+        adolc_exit(-2,"",__func__,__FILE__,__LINE__);
     };
 
     if((numdep != depen)||(numind != indep)) {
@@ -563,7 +566,7 @@ int hov_ti_reverse(
                 "the number of\nindependent and/or dependent variables"
                 " given to reverse are\ninconsistent with that of the"
                 "  internal taylor array.\n",tnum);
-        exit(-2);
+        adolc_exit(-2,"",__func__,__FILE__,__LINE__);
     }
 
 
@@ -1266,6 +1269,7 @@ int hov_ti_reverse(
                   }
             }
 
+		if (arg != res)
                 GET_TAYL(res,k,p)
                 break;
 
@@ -1573,8 +1577,8 @@ int hov_ti_reverse(
                                 if (coval == 1) {
                                     FOR_0_LE_i_LT_k
                                     { /* ! no tempory */
-                                        AARG_INC += ARES;
-                                        ARES_INC = 0.0;
+                                        Aarg[i] += Ares[i];
+                                        Ares[i] = 0.0;
                                     }
                                 } else {
                                     /* coval is an int > 1 */
@@ -2066,7 +2070,7 @@ int hov_ti_reverse(
 	    case subscript:
 		coval = get_val_r();
 		{
-		    size_t cnt, idx, numval = (size_t)trunc(fabs(coval));
+		    size_t idx, numval = (size_t)trunc(fabs(coval));
 		    locint vectorloc;
 		    vectorloc = get_locint_r();
 		    res = get_locint_r();
@@ -2074,7 +2078,7 @@ int hov_ti_reverse(
 		    ASSIGN_T(Targ, rpp_T[arg])
 		    idx = (size_t)trunc(fabs(TARG));
 		    if (idx >= numval)
-			fprintf(DIAG_OUT, "ADOL-C warning: index out of bounds while subscripting n=%z, idx=%z\n", numval, idx);
+			fprintf(DIAG_OUT, "ADOL-C warning: index out of bounds while subscripting n=%zu, idx=%zu\n", numval, idx);
 		    arg1 = vectorloc+idx;
 		    ASSIGN_A(Aarg1, rpp_A[arg1])
 		    ASSIGN_A(Ares, rpp_A[res])
@@ -2100,7 +2104,7 @@ int hov_ti_reverse(
 	    case subscript_ref:
 		coval = get_val_r();
 		{
-		    size_t cnt, idx, numval = (size_t)trunc(fabs(coval));
+		    size_t idx, numval = (size_t)trunc(fabs(coval));
 		    locint vectorloc;
 		    vectorloc = get_locint_r();
 		    res = get_locint_r();
@@ -2109,14 +2113,16 @@ int hov_ti_reverse(
 		    ASSIGN_T(Tres, rpp_T[res])
 		    idx = (size_t)trunc(fabs(TARG));
 		    if (idx >= numval)
-			fprintf(DIAG_OUT, "ADOL-C warning: index out of bounds while subscripting (ref) n=%z, idx=%z\n", numval, idx);
+			fprintf(DIAG_OUT, "ADOL-C warning: index out of bounds while subscripting (ref) n=%zu, idx=%zu\n", numval, idx);
 		    arg1 = (size_t)trunc(fabs(TRES));
-		    // This is actually NOP 
-                    // basically all we need is that arg1 == vectorloc[idx]
-                    // so doing a check here is probably good
+		    /*
+		     * This is actually NOP
+                     * basically all we need is that arg1 == vectorloc[idx]
+                     * so doing a check here is probably good
+                     */
 		    if (arg1 != vectorloc+idx) {
-			fprintf(DIAG_OUT, "ADOL-C error: indexed active position does not match referenced position\nindexed = %d, referenced = %d\n", vectorloc+idx, arg1);
-			exit(-2);
+			fprintf(DIAG_OUT, "ADOL-C error: indexed active position does not match referenced position\nindexed = %zu, referenced = %d\n", vectorloc+idx, arg1);
+			adolc_exit(-2,"",__func__,__FILE__,__LINE__);
 		    }
 		    GET_TAYL(res,k,p)
 		}
@@ -2162,7 +2168,7 @@ int hov_ti_reverse(
             case ref_assign_d:      /* assign an adouble variable a    assign_d */
                 /* double value. (=) */
                 coval = get_val_r();
-		// fallthrough 
+		/* fallthrough */
             case ref_assign_d_zero: /* assign an adouble a        assign_d_zero */
             case ref_assign_d_one:  /* double value. (=)           assign_d_one */
                 arg1   = get_locint_r();
@@ -2513,7 +2519,7 @@ int hov_ti_reverse(
             case take_stock_op:                                  /* take_stock_op */
                 res = get_locint_r();
                 size = get_locint_r();
-                d = get_val_v_r(size);
+                get_val_v_r(size);
 
                 res += size;
                 for (ls=size;ls>0;ls--) {
@@ -2551,7 +2557,7 @@ int hov_ti_reverse(
                 fprintf(DIAG_OUT,"ADOL-C fatal error in " GENERATED_FILENAME " ("
                         __FILE__
                         ") : no such operation %d\n", operation);
-                exit(-1);
+                adolc_exit(-1,"",__func__,__FILE__,__LINE__);
                 break;
         } /* endswitch */
 
@@ -2583,6 +2589,7 @@ int hov_ti_reverse(
     free((char*) jj);
     free((char*) x);
 
+    ADOLC_CURRENT_TAPE_INFOS.workMode = ADOLC_NO_MODE;
     end_sweep();
 
     return ret_c;
