@@ -1,7 +1,7 @@
 /*----------------------------------------------------------------------------
  ADOL-C -- Automatic Differentiation by Overloading in C++
  File:     taping.c
- Revision: $Id: taping.c 370 2012-11-22 13:18:52Z kulshres $
+ Revision: $Id: taping.c 527 2014-07-15 14:09:31Z kulshres $
  Contents: all C functions directly accessing at least one of the four tapes
            (operations, locations, constants, value stack)
 
@@ -24,6 +24,11 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 
+#ifdef ADOLC_AMPI_SUPPORT
+#include "ampi/ampi.h"
+#include "ampi/tape/support.h"
+#endif
+
 #if defined(_WINDOWS) && !__STDC__
 #define stat _stat
 #define S_IFDIR _S_IFDIR
@@ -180,7 +185,7 @@ void fail( int error ) {
                     "independents(%u)\n"
                     "              variables passed to reverse is "
                     "inconsistent\n"
-                    "              with number recorded on tape(%u/%u)!\n",
+                    "              with number recorded on tape(%zu/%zu)!\n",
                     ADOLC_CURRENT_TAPE_INFOS.tapeID, failAdditionalInfo3,
                     failAdditionalInfo4,
                     ADOLC_CURRENT_TAPE_INFOS.stats[NUM_DEPENDENTS],
@@ -279,13 +284,19 @@ void fail( int error ) {
 		    "ADOL-C error: Trace was created on a 32-bit platform, cannot be opened on 64-bit platform!\n"
 		);
 	    break;
+        case ADOLC_TAPING_NOT_ACTUALLY_TAPING:
+	    fprintf(DIAG_OUT,
+		    "ADOL-C error: Trace %d is not being currently created!\n",
+		    failAdditionalInfo1);
+	    break;
 
         default:
             fprintf(DIAG_OUT, "ADOL-C error => unknown error type!\n");
-            exit (-1);
+            adolc_exit(-1, "", __func__, __FILE__, __LINE__);
             break;
     }
-    exit (error + 1);
+    adolc_exit(error+1, "", __func__,  __FILE__, __LINE__);
+    // exit (error + 1);
 }
 
 /* print an error message describing the error number */
@@ -328,18 +339,22 @@ void printError() {
 }
 
 /* the base names of every tape type */
-char *tapeBaseNames[4];
+char *tapeBaseNames[4]={0,0,0,0};
 
 void clearTapeBaseNames() {
     int i;
-    for(i=0;i<4;i++)
-	free(tapeBaseNames[i]);
+    for(i=0;i<4;i++) {
+	if (tapeBaseNames[i]) {
+	    free(tapeBaseNames[i]);
+	    tapeBaseNames[i]=0;
+	}
+    }
 }
 
 /****************************************************************************/
 /* The subroutine get_fstr appends to the tape base name of type tapeType   */
 /* the number fnum and ".tap" and returns a pointer to the resulting string.*/
-/* The result string must be freed be thy caller!                           */
+/* The result string must be freed be the caller!                           */
 /****************************************************************************/
 char *createFileName(short tapeID, int tapeType) {
     char *numberString, *fileName, *extension = ".tap", *currPos;
@@ -417,6 +432,13 @@ char *createFileName(short tapeID, int tapeType) {
 /****************************************************************************/
 /* Tries to read a local config file containing, e.g., buffer sizes         */
 /****************************************************************************/
+static char* duplicatestr(const char* instr) {
+    size_t len = strlen(instr);
+    char *outstr = calloc(len+1,sizeof(char));
+    strncpy(outstr,instr,len);
+    return outstr;
+}
+
 #define ADOLC_LINE_LENGTH 100
 void readConfigFile() {
     FILE *configFile = NULL;
@@ -426,13 +448,13 @@ void readConfigFile() {
     long int number = 0;
     char *path = NULL;
     int defdirsize = strlen(TAPE_DIR PATHSEPARATOR);
-    tapeBaseNames[0] = strdup(
+    tapeBaseNames[0] = duplicatestr(
 	TAPE_DIR PATHSEPARATOR ADOLC_LOCATIONS_NAME);
-    tapeBaseNames[1] = strdup(
+    tapeBaseNames[1] = duplicatestr(
 	TAPE_DIR PATHSEPARATOR ADOLC_VALUES_NAME);
-    tapeBaseNames[2] = strdup(
+    tapeBaseNames[2] = duplicatestr(
 	TAPE_DIR PATHSEPARATOR ADOLC_OPERATIONS_NAME);
-    tapeBaseNames[3] = strdup(
+    tapeBaseNames[3] = duplicatestr(
 	TAPE_DIR PATHSEPARATOR ADOLC_TAYLORS_NAME);
 
     ADOLC_OPENMP_THREAD_NUMBER;
@@ -623,11 +645,11 @@ void take_stock() {
 /* - intended to be used in stop_trace only                                 */
 /****************************************************************************/
 locint keep_stock() {
+    locint loc2;
     ADOLC_OPENMP_THREAD_NUMBER;
     ADOLC_OPENMP_GET_THREAD_NUMBER;
-    /* if we have adoubles in use */
-    if (ADOLC_GLOBAL_TAPE_VARS.numLives > 0) {
-        locint loc2 = ADOLC_GLOBAL_TAPE_VARS.storeSize - 1;
+    /* save all the final adoubles when finishing tracing */
+        loc2 = ADOLC_GLOBAL_TAPE_VARS.storeSize - 1;
 
         /* special signal -> all alive adoubles recorded on the end of the
          * value stack -> special handling at the beginning of reverse */
@@ -642,7 +664,6 @@ locint keep_stock() {
                 ADOLC_WRITE_SCAYLOR(ADOLC_GLOBAL_TAPE_VARS.store[loc2]);
             } while (loc2-- > 0);
         }
-    }
     ADOLC_CURRENT_TAPE_INFOS.traceFlag = 0;
     return ADOLC_GLOBAL_TAPE_VARS.storeSize;
 }
@@ -651,7 +672,7 @@ locint keep_stock() {
 /****************************************************************************/
 /* Set up statics for writing taylor data                                   */
 /****************************************************************************/
-void taylor_begin(uint bufferSize, double **Tg, int degreeSave) {
+void taylor_begin(uint bufferSize, int degreeSave) {
     ADOLC_OPENMP_THREAD_NUMBER;
     ADOLC_OPENMP_GET_THREAD_NUMBER;
     if (ADOLC_CURRENT_TAPE_INFOS.tayBuffer != NULL) {
@@ -671,7 +692,6 @@ void taylor_begin(uint bufferSize, double **Tg, int degreeSave) {
     }
 
     /* initial setups */
-    ADOLC_CURRENT_TAPE_INFOS.dpp_T = Tg;
     if (ADOLC_CURRENT_TAPE_INFOS.tayBuffer != NULL)
         free(ADOLC_CURRENT_TAPE_INFOS.tayBuffer);
     ADOLC_CURRENT_TAPE_INFOS.tayBuffer = (revreal *)
@@ -972,6 +992,8 @@ void get_taylors_p(locint loc, int degree, int numDir) {
         --T; /* skip the base point part */
     }
     /* now update the base point parts */
+    if (ADOLC_CURRENT_TAPE_INFOS.currTay == ADOLC_CURRENT_TAPE_INFOS.tayBuffer)
+	get_tay_block_r();
     --ADOLC_CURRENT_TAPE_INFOS.currTay;
     for (i = 0; i < numDir; ++i) {
         *T = *ADOLC_CURRENT_TAPE_INFOS.currTay;
@@ -1059,6 +1081,8 @@ void start_trace() {
     ADOLC_CURRENT_TAPE_INFOS.currLoc = ADOLC_CURRENT_TAPE_INFOS.locBuffer;
     ADOLC_CURRENT_TAPE_INFOS.currVal = ADOLC_CURRENT_TAPE_INFOS.valBuffer;
     ADOLC_CURRENT_TAPE_INFOS.num_eq_prod = 0;
+    ADOLC_CURRENT_TAPE_INFOS.numSwitches = 0;
+    ADOLC_CURRENT_TAPE_INFOS.workMode = ADOLC_TAPING;
 
     /* Put operation denoting the start_of_the tape */
     put_op(start_of_tape);
@@ -1071,7 +1095,7 @@ void start_trace() {
 
     /* initialize value stack if necessary */
     if (ADOLC_CURRENT_TAPE_INFOS.keepTaylors)
-        taylor_begin(ADOLC_CURRENT_TAPE_INFOS.stats[TAY_BUFFER_SIZE], NULL, 0);
+        taylor_begin(ADOLC_CURRENT_TAPE_INFOS.stats[TAY_BUFFER_SIZE], 0);
 
     /* mark possible (hard disk) tape creation */
     markNewTape();
@@ -1094,6 +1118,10 @@ void stop_trace(int flag) {
 
     ADOLC_CURRENT_TAPE_INFOS.stats[NUM_EQ_PROD] = 
         ADOLC_CURRENT_TAPE_INFOS.num_eq_prod; 
+
+    ADOLC_CURRENT_TAPE_INFOS.stats[NUM_SWITCHES] =
+	ADOLC_CURRENT_TAPE_INFOS.numSwitches;
+
     taylor_close(ADOLC_CURRENT_TAPE_INFOS.stats[TAY_BUFFER_SIZE]);
 
     /* The taylor stack size base estimation results in a doubled taylor count
@@ -1323,6 +1351,11 @@ void read_tape_stats(TapeInfos *tapeInfos) {
     tapeInfos->tapingComplete = 1;
 }
 
+void skip_tracefile_cleanup(short tnum) {
+    TapeInfos *tinfo = getTapeInfos(tnum);
+    tinfo->pTapeInfos.skipFileCleanup = 1;
+}
+
 /****************************************************************************/
 /* Initialize a forward sweep. Get stats, open tapes, fill buffers, ...     */
 /****************************************************************************/
@@ -1435,6 +1468,9 @@ void init_for_sweep(short tag) {
     }
     ADOLC_CURRENT_TAPE_INFOS.numVals_Tape = number;
     ADOLC_CURRENT_TAPE_INFOS.currVal = ADOLC_CURRENT_TAPE_INFOS.valBuffer;
+#ifdef ADOLC_AMPI_SUPPORT
+    TAPE_AMPI_resetBottom();
+#endif
 }
 
 /****************************************************************************/
@@ -1552,6 +1588,9 @@ void init_rev_sweep(short tag) {
         ADOLC_CURRENT_TAPE_INFOS.stats[NUM_VALUES] - number;
     ADOLC_CURRENT_TAPE_INFOS.currVal =
         ADOLC_CURRENT_TAPE_INFOS.valBuffer + number;
+#ifdef ADOLC_AMPI_SUPPORT
+    TAPE_AMPI_resetTop();
+#endif
 }
 
 /****************************************************************************/
@@ -1578,17 +1617,22 @@ void end_sweep() {
 
 /* --- Operations --- */
 
+#if defined(__USE_ISOC99)
+const int maxLocsPerOp=10;
+#endif
+
 /****************************************************************************/
 /* Puts an operation into the operation buffer. Ensures that location buffer*/
 /* and constants buffer are prepared to take the belonging stuff.           */
 /****************************************************************************/
-void put_op(unsigned char op) {
+void put_op_reserve(unsigned char op, unsigned int reserveExtraLocations) {
     ADOLC_OPENMP_THREAD_NUMBER;
     ADOLC_OPENMP_GET_THREAD_NUMBER;
-    /* every operation writes <5 locations */
-    if (ADOLC_CURRENT_TAPE_INFOS.currLoc + 5 > ADOLC_CURRENT_TAPE_INFOS.lastLocP1) {
-        *(ADOLC_CURRENT_TAPE_INFOS.lastLocP1 - 1) = ADOLC_CURRENT_TAPE_INFOS.lastLocP1 -
-                ADOLC_CURRENT_TAPE_INFOS.currLoc;
+    /* make sure we have enough slots to write the locs */
+    if (ADOLC_CURRENT_TAPE_INFOS.currLoc + maxLocsPerOp + reserveExtraLocations > ADOLC_CURRENT_TAPE_INFOS.lastLocP1) {
+        size_t remainder = ADOLC_CURRENT_TAPE_INFOS.lastLocP1 - ADOLC_CURRENT_TAPE_INFOS.currLoc;
+        if (remainder>0) memset(ADOLC_CURRENT_TAPE_INFOS.currLoc,0,(remainder-1)*sizeof(locint));
+        *(ADOLC_CURRENT_TAPE_INFOS.lastLocP1 - 1) = remainder;
         put_loc_block(ADOLC_CURRENT_TAPE_INFOS.lastLocP1);
         /* every operation writes 1 opcode */
         if (ADOLC_CURRENT_TAPE_INFOS.currOp + 1 == ADOLC_CURRENT_TAPE_INFOS.lastOpP1) {
@@ -1602,7 +1646,10 @@ void put_op(unsigned char op) {
     }
     /* every operation writes <5 values --- 3 should be sufficient */
     if (ADOLC_CURRENT_TAPE_INFOS.currVal + 5 > ADOLC_CURRENT_TAPE_INFOS.lastValP1) {
-        ADOLC_PUT_LOCINT(ADOLC_CURRENT_TAPE_INFOS.lastValP1 - ADOLC_CURRENT_TAPE_INFOS.currVal);
+        locint valRemainder=ADOLC_CURRENT_TAPE_INFOS.lastValP1 - ADOLC_CURRENT_TAPE_INFOS.currVal;
+        ADOLC_PUT_LOCINT(valRemainder);
+        /* avoid writing uninitialized memory to the file and get valgrind upset */
+        memset(ADOLC_CURRENT_TAPE_INFOS.currVal,0,valRemainder*sizeof(double));
         put_val_block(ADOLC_CURRENT_TAPE_INFOS.lastValP1);
         /* every operation writes 1 opcode */
         if (ADOLC_CURRENT_TAPE_INFOS.currOp + 1 == ADOLC_CURRENT_TAPE_INFOS.lastOpP1) {
